/*
 * $Header: /home/cvs/jakarta-slide/src/util/org/apache/util/HttpsURL.java,v 1.3.2.1 2004/02/05 16:07:54 mholz Exp $
 * $Revision: 1.3.2.1 $
 * $Date: 2004/02/05 16:07:54 $
 *
 * ====================================================================
 *
 * Copyright 1999-2002 The Apache Software Foundation 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */ 

package org.apache.util;

/**
 * This is for the HTTPS URL of the generic URI version.
 *
 * @author <a href="mailto:jericho@apache.org">Sung-Gu</a>
 */
public class HttpsURL extends HttpURL {

    // --------------------------------------------------------- Constructors


    /**
     * This Constructor
     *
     * @param escapedHttpsURL The escaped https URL string.
     */
    public HttpsURL(String escapedHttpsURL) {
        super(escapedHttpsURL);
        setDefaultScheme(scheme);
        setDefaultPort(port);
    }


    /**
     * This Constructor
     *
     * @param host The host string.
     * @param path The path string.
     */
    public HttpsURL(String host, String path) {
        this(null, null, host, HttpsURL.port, path, null);
    }


    /**
     * This Constructor
     *
     * @param host The host string.
     * @param port The port number.
     * @param path The path string.
     */
    public HttpsURL(String host, int port, String path) {
        this(null, null, host, port, path, null);
    }


    /**
     * This Constructor
     *
     * @param host The host string.
     * @param port The port number.
     * @param path The path string.
     * @param query The query string.
     */
    public HttpsURL(String host, int port, String path, String query) {
        this(null, null, host, port, path, query);
    }


    /**
     * This Constructor
     *
     * @param userName The username string.
     * @param password The password string.
     * @param host The host string.
     */
    public HttpsURL(String userName, String password, String host) {
        this(userName, password, host, HttpsURL.port, null, null);
    }


    /**
     * This Constructor
     *
     * @param userName The username string.
     * @param password The password string.
     * @param host The host string.
     * @param port The port number.
     */
    public HttpsURL(String userName, String password, String host, int port) {
        this(userName, password, host, port, null, null);
    }


    /**
     * This Constructor
     *
     * @param userName The username string.
     * @param password The password string.
     * @param host The host string.
     * @param path The path string.
     */
    public HttpsURL(String userName, String password, String host,
                   String path) {
        this(userName, password, host, HttpsURL.port, path, null);
    }


    /**
     * This Constructor
     *
     * @param userName The username string.
     * @param password The password string.
     * @param host The host string.
     * @param port The port number.
     * @param path The path string.
     */
    public HttpsURL(String userName, String password, String host, int port,
                   String path) {
        this(userName, password, host, port, path, null);
    }


    /**
     * This Constructor
     *
     * @param userName The username string.
     * @param password The password string.
     * @param host The host string.
     * @param path The path string.
     * @param query The query string.
     */
    public HttpsURL(String userName, String password, String host,
                   String path, String query) {
        this(userName, password, host, HttpsURL.port, path, query);
    }


    /**
     * This Constructor
     *
     * @param userName The username string.
     * @param password The password string.
     * @param host The host string.
     * @param port The port number.
     * @param path The path string.
     * @param query The query string.
     */
    public HttpsURL(String userName, String password, String host, int port,
                   String path, String query) {

        super
            (URIUtil.escape(HttpsURL.scheme, URIUtil.schemeReserved()) + "://" +
             ((userName == null) ? "" :
              URIUtil.escape(userName) + ((password == null) ? "" :
                                          ":" + URIUtil.escape(password))
              + "@") +
             URIUtil.escape(host, URIUtil.hostReserved()) +
             ((port == HttpsURL.port || port == -1) ? "" : ":" + port) +
             ((path == null || path.equals("")) ? "/" :
              URIUtil.escape((!path.startsWith("/")) ?
                             "/" + path : path, URIUtil.pathReserved())) +
             ((query == null || query.equals("")) ? "" :
              "?" + URIUtil.escape(query, URIUtil.queryReserved())));
        setDefaultScheme(scheme);
        setDefaultPort(port);
    }


    /**
     * This Constructor
     *
     * @param httpsURL The https URL.
     * @param path The added relative path.
     */
    public HttpsURL(HttpsURL httpsURL, String path) {

        this(httpsURL.toString() +
             (httpsURL.toString().endsWith("/") ? "" : "/") +
             URIUtil.escape(path, URIUtil.pathReserved()));
    }
    

    // ------------------------------------------------------------ Constants


    /**
     * Default scheme for HTTPS URL.
     */
    public final static String scheme = "https";


    /**
     * Default port for HTTPS URL.
     */
    public final static int port = 443;


    /**
     * The serialVersionUID.
     */
    static final long serialVersionUID = 887844277028676647L;


    // ------------------------------------------------------  Public Methods

}
